/*

[The "BSD licence"]
Copyright (c) 2005 Jean Bovet
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

package org.antlr.xjlib.appkit.frame;

import org.antlr.xjlib.appkit.XJControl;
import org.antlr.xjlib.appkit.app.XJApplication;
import org.antlr.xjlib.appkit.app.XJPreferences;
import org.antlr.xjlib.appkit.menu.*;
import org.antlr.xjlib.appkit.undo.XJUndo;
import org.antlr.xjlib.appkit.undo.XJUndoDelegate;
import org.antlr.xjlib.appkit.undo.XJUndoEngine;
import org.antlr.xjlib.appkit.utils.XJLocalizable;
import org.antlr.xjlib.foundation.XJSystem;

import java.util.List;
import javax.swing.*;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;
import javax.swing.text.DefaultEditorKit;
import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyVetoException;
import java.util.HashMap;
import java.util.Map;
import java.util.ArrayList;

public class XJFrame extends XJControl implements XJFrameInterface, XJMenuBarCustomizer, XJMenuBarDelegate {

    private static final String PROPERTY_WINDOW_MODIFIED = "windowModified";
    private static final String PREF_DESKTOP_BOUNDS = "xjdesktop_bounds";

    private static final boolean useDesktop = XJApplication.shared().useDesktopMode();

    private static JFrame desktopFrame;
    private static JDesktopPane desktop;
    private static XJMainMenuBar desktopDefaultMenuBar;

    private JInternalFrame jInternalFrame;
    private JFrame jFrame;

    protected XJMainMenuBar mainMenuBar;
    protected XJFrameDelegate delegate;
    protected XJUndoEngine undoEngine;
    protected boolean alreadyBecomeVisible = false;
    protected boolean dirty = false;

    private class InternalFrameHandling implements XJMenuBarDelegate {

        public InternalFrameHandling() {
            desktopDefaultMenuBar = XJMainMenuBar.createInstance();
            desktopDefaultMenuBar.createMenuBar();
            desktopDefaultMenuBar.setDelegate(this);
            desktopDefaultMenuBar.refreshState();
        }

        public void menuItemState(XJMenuItem item) {
            int tag = item.getTag();
            switch(tag) {
                case XJMainMenuBar.MI_NEW:
                    item.setTitle(XJLocalizable.getXJString("New")+((XJApplication.shared().getDocumentExtensions().size()>1)?"...":""));

                case XJMainMenuBar.MI_OPEN:
                case XJMainMenuBar.MI_RECENT_FILES:
                case XJMainMenuBar.MI_CLEAR_RECENT_FILES:
                    item.setEnabled(true);
                    break;

                default:
                    if(XJMainMenuBar.isRecentFilesItem(item))
                        item.setEnabled(true);
                    else
                        item.setEnabled(false);
                    break;
            }
        }

        public void handleMenuEvent(XJMenu menu, XJMenuItem item) {

        }

        public void handleMenuSelected(XJMenu menu) {

        }
    }

    private static class FrameManager {

        private static List<XJFrame> frames = new ArrayList<XJFrame>();
        private static XJFrame activated = null;

        public static void removeFrame(XJFrame f) {
            frames.remove(f);
        }

        public static void frameActivated(XJFrame f) {
            frames.remove(f);
            frames.add(f);
            activated = f;
        }

        public static void frameDeactivated(final XJFrame f) {
            if(activated == f) {
                activated = null;
            }

            int index = frames.indexOf(f);
            if(index == -1) return;

            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    selectNextFrame(f);
                }
            });
        }

        public static XJFrame getNextFrame(XJFrame f) {
            for(int i=frames.size()-1; i>=0; i--) {
                if(frames.get(i).equals(f)) continue;
                return frames.get(i);
            }
            return null;
        }

        public static void selectNextFrame(XJFrame f) {
            if(XJSystem.isMacOS()) return;

            if(activated == f) return;

            XJFrame nf = getNextFrame(f);
            if(nf != null) {
                nf.bringToFront();
            } else if(frames.isEmpty() || frames.size() == 1 && frames.get(0).equals(f) && desktopFrame != null) {
                desktopFrame.setJMenuBar(desktopDefaultMenuBar.getJMenuBar());
                desktopFrame.toFront();
            }
        }
    }

    private static void restoreDesktopBounds() {
        Rectangle r = (Rectangle) XJApplication.shared().getPreferences().getObject(PREF_DESKTOP_BOUNDS, null);
        if(r != null) {
            desktopFrame.setLocation(r.x, r.y);
            desktopFrame.setSize(r.width, r.height);
        }
    }

    private static void saveDesktopBounds() {
        Point pos = desktopFrame.getLocation();
        Dimension s = desktopFrame.getSize();
        Rectangle r = new Rectangle(pos.x, pos.y, s.width, s.height);
        XJPreferences prefs = XJApplication.shared().getPreferences();
        prefs.setObject(PREF_DESKTOP_BOUNDS, r);
    }

    public static void closeDesktop() {
        if(useDesktop) {
            saveDesktopBounds();
        }
    }

    public XJFrame() {
        if(useDesktop) {
            if(desktopFrame == null) {
                desktopFrame = new JFrame();
                desktopFrame.setTitle(XJApplication.shared().getApplicationName());
                desktopFrame.addWindowListener(new WindowAdapter() {
                    public void windowClosing(WindowEvent e) {
                        XJApplication.shared().performQuit();
                    }

                });

                Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
                dim = new Dimension((int)(dim.width*0.9), (int)(dim.height*0.9));
                desktopFrame.setSize(dim);
                desktopFrame.setPreferredSize(dim);
                desktopFrame.setLocationRelativeTo(null);
                restoreDesktopBounds();

                desktop = new JDesktopPane();
                desktopFrame.getContentPane().add(desktop, BorderLayout.CENTER);
                desktopFrame.setVisible(true);

                new InternalFrameHandling();
            }

            jInternalFrame = new JInternalFrame();
            jInternalFrame.setResizable(true);
            jInternalFrame.setClosable(true);
            jInternalFrame.setMaximizable(true);
            jInternalFrame.setIconifiable(true);

            desktop.add(jInternalFrame);

            jInternalFrame.addInternalFrameListener(new InternalFrameAdapter() {
                public void internalFrameActivated(InternalFrameEvent e) {
                    XJMainMenuBar mb = getMainMenuBar();
                    if(mb != null) {
                        desktopFrame.setJMenuBar(mb.getJMenuBar());
                    }
                    XJFrame.this.windowActivated();
                    FrameManager.frameActivated(XJFrame.this);
                }

                public void internalFrameDeactivated(InternalFrameEvent e) {
                    // Don't send the deactivate event if the frame is closed
                    // (because this event is send also when the window just closed
                    // which is kind of weird)
                    if(jInternalFrame != null)
                        XJFrame.this.windowDeactivated();

                    FrameManager.frameDeactivated(XJFrame.this);
                }

                public void internalFrameClosing(InternalFrameEvent e) {
                    if(desktopFrame != null && jInternalFrame.isSelected()) {
                        desktopFrame.setJMenuBar(desktopDefaultMenuBar.getJMenuBar());
                    }
                    XJFrame.this.windowClosing();
                    FrameManager.removeFrame(XJFrame.this);
                }
            });
        } else {
            jFrame = new JFrame();
            jFrame.addWindowListener(new WindowAdapter() {
                public void windowActivated(WindowEvent e) {
                    XJFrame.this.windowActivated();
                    FrameManager.frameActivated(XJFrame.this);
                }

                public void windowDeactivated(WindowEvent e) {
                    // Don't send the deactivate event if the frame is closed
                    // (because this event is send also when the window just closed
                    // which is kind of weird)
                    if(jFrame != null)
                        XJFrame.this.windowDeactivated();
                    FrameManager.frameDeactivated(XJFrame.this);
                }

                public void windowClosing(WindowEvent e) {
                    XJFrame.this.windowClosing();
                    FrameManager.removeFrame(XJFrame.this);
                }

            });
        }
        setDefaultSize();
        undoEngine = new XJUndoEngine();
    }

    public void awake() {
        if(shouldDisplayMainMenuBar()) {
            mainMenuBar = XJMainMenuBar.createInstance();
            mainMenuBar.setCustomizer(this);
            mainMenuBar.setDelegate(this);
            mainMenuBar.createMenuBar();
            setMainMenuBar(mainMenuBar);
            undoEngine.setMainMenuBar(mainMenuBar);
        }
    }

    public void setDelegate(XJFrameDelegate delegate) {
        this.delegate = delegate;
    }

    public XJFrameDelegate getDelegate() {
        return delegate;
    }

    public void setDefaultCloseOperation(int operation) {
        if(useDesktop) {
            jInternalFrame.setDefaultCloseOperation(operation);
        } else {
            jFrame.setDefaultCloseOperation(operation);
        }
    }

    public Container getContentPane() {
        if(useDesktop) {
            return jInternalFrame.getContentPane();
        } else {
            return jFrame.getContentPane();
        }
    }

    public JRootPane getRootPane() {
        if(useDesktop) {
            return jInternalFrame.getRootPane();
        } else {
            return jFrame.getRootPane();
        }
    }

    public JLayeredPane getLayeredPane() {
        if(useDesktop) {
            return jInternalFrame.getLayeredPane();
        } else {
            return jFrame.getLayeredPane();
        }
    }

    public Component getGlassPane() {
        if(useDesktop) {
            return jInternalFrame.getGlassPane();
        } else {
            return jFrame.getGlassPane();
        }
    }

    public void setMainMenuBar(XJMainMenuBar menubar) {
        this.mainMenuBar = menubar;
        if(useDesktop) {
            if(jInternalFrame.isSelected()) {
                desktopFrame.setJMenuBar(mainMenuBar.getJMenuBar());
            }
        } else {
            jFrame.setJMenuBar(mainMenuBar.getJMenuBar());
        }
    }

    public XJMainMenuBar getMainMenuBar() {
        return mainMenuBar;
    }

    public void menuItemStatusChanged(int tag) {
        if(mainMenuBar == null)
            return;

        XJMainMenuBar.refreshAllMenuBars();
        mainMenuBar.refreshState();
    }

    public void setTitle(String title) {
        if(useDesktop) {
            jInternalFrame.setTitle(title);
        } else {
            jFrame.setTitle(title);
        }
    }

    public String getTitle() {
        if(useDesktop) {
            return jInternalFrame.getTitle();
        } else {
            return jFrame.getTitle();
        }
    }

    public void setLocation(Point loc) {
        if(useDesktop) {
            jInternalFrame.setLocation(loc);
        } else {
            jFrame.setLocation(loc);
        }
    }

    public Point getLocation() {
        if(useDesktop) {
            return jInternalFrame.getLocation();
        } else {
            return jFrame.getLocation();
        }
    }

    public void setSize(int dx, int dy) {
        if(useDesktop) {
            jInternalFrame.setSize(dx, dy);
        } else {
            jFrame.setSize(dx, dy);
        }
    }

    public void setSize(Dimension size) {
        if(useDesktop) {
            jInternalFrame.setSize(size);
        } else {
            jFrame.setSize(size);
        }
    }

    public Dimension getSize() {
        if(useDesktop) {
            return jInternalFrame.getSize();
        } else {
            return jFrame.getSize();
        }
    }

    public void setDefaultSize() {
        if(useDesktop) {
            Dimension dim = desktop.getSize();
            setSize((int) (dim.width*0.8), (int) (dim.height*0.8));
        } else {
            Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
            setSize((int)(dim.width*0.5), (int)(dim.height*0.5));
        }
    }

    public void setResizable(boolean flag) {
        if(useDesktop) {
            jInternalFrame.setResizable(flag);
        } else {
            jFrame.setResizable(flag);
        }
    }

    public void setMaximizable(boolean flag) {
        if(useDesktop) {
            jInternalFrame.setMaximizable(flag);
        } else {
            // not applicable
        }
    }

    public boolean isMaximized() {
        if(useDesktop) {
            return jInternalFrame.isMaximum();
        } else {
            return false;
        }
    }

    public void pack() {
        if(useDesktop) {
            jInternalFrame.pack();
        } else {
            jFrame.pack();
        }
    }

    public void bringToFront() {
        if(useDesktop) {
            jInternalFrame.moveToFront();
            try {
                jInternalFrame.setSelected(true);
            } catch (PropertyVetoException e) {
                e.printStackTrace();
            }
        } else {
            jFrame.toFront();
        }
    }

    public void setVisible(boolean flag) {
        if(flag && !alreadyBecomeVisible) {
            alreadyBecomeVisible = true;
            restoreWindowBounds();
            ensureVisibility();
            becomingVisibleForTheFirstTime();
        }
        if(useDesktop) {
            jInternalFrame.setVisible(flag);
        } else {
            jFrame.setVisible(flag);
        }
        bringToFront();
    }

    /**
     * Ensures that the window is visible in one of the available screen.
     * Otherwise, move it to the center of the main screen.
     */
    public void ensureVisibility() {
        if(!isVisibleOnScreen()) {
            center();
        }
    }

    public void becomingVisibleForTheFirstTime() {

    }

    public String autosaveName() {
        return null;
    }

    public boolean isVisible() {
        if(useDesktop) {
            return jInternalFrame.isVisible();
        } else {
            return jFrame.isVisible();
        }
    }

    public boolean isActive() {
        if(useDesktop) {
            return jInternalFrame.isSelected();
        } else {
            return jFrame.isActive();
        }
    }

    public void show() {
        setVisible(true);
    }

    public void showModal() {
        setVisible(true);
    }

    public void hide() {
        setVisible(false);
    }

    public boolean isCompletelyOnScreen() {
        return isVisibleOnScreen();
    }

    public boolean isVisibleOnScreen() {
        GraphicsEnvironment ge =
                GraphicsEnvironment.getLocalGraphicsEnvironment();
        GraphicsDevice[] gs = ge.getScreenDevices();

        Rectangle fr = (useDesktop?jInternalFrame.getBounds():jFrame.getBounds());
        for (GraphicsDevice g : gs) {
            GraphicsConfiguration dc = g.getDefaultConfiguration();
            if (fr.intersects(dc.getBounds())) {
                return true;
            }
        }

        return false;
    }

    public void center() {
        if(useDesktop) {
            Dimension s = desktopFrame.getSize();
            Dimension id = jInternalFrame.getSize();
            jInternalFrame.setLocation(s.width/2 - id.width/2, s.height/2 - id.height/2);
        } else {
            jFrame.setLocationRelativeTo(null);
        }
    }

    public void setPosition(int x, int y) {
        setLocation(new Point(x, y));
    }

    public void offsetPosition(int dx, int dy) {
        Point p = getLocation();
        setPosition(p.x+dx, p.y+dy);
    }

    public void close() {
        XJMainMenuBar.removeInstance(mainMenuBar);
        if(mainMenuBar != null) {
            mainMenuBar.setDelegate(null);
            mainMenuBar = null;
        }

        saveWindowBounds();

        if(useDesktop) {
            jInternalFrame.dispose();
            jInternalFrame = null;
        } else {
            jFrame.dispose();
            jFrame = null;
        }

        if(delegate != null)
            delegate.frameDidClose(this);
    }

    public void setDirty() {
        if(!XJApplication.shared().supportsPersistence())
            return;

        // Use dirty member to speed up
        if(!dirty) {
            dirty = true;
            getRootPane().putClientProperty(PROPERTY_WINDOW_MODIFIED, Boolean.TRUE);
            menuItemStatusChanged(XJMainMenuBar.MI_SAVE);
        }
    }

    public void resetDirty() {
        if(!XJApplication.shared().supportsPersistence())
            return;

        if(dirty) {
            dirty = false;
            getRootPane().putClientProperty(PROPERTY_WINDOW_MODIFIED, Boolean.FALSE);
            menuItemStatusChanged(XJMainMenuBar.MI_SAVE);
        }
    }

    public boolean dirty() {
        if(!XJApplication.shared().supportsPersistence())
            return false;

        Boolean b = (Boolean) getRootPane().getClientProperty(PROPERTY_WINDOW_MODIFIED);
        if(b == null)
            return false;
        else
            return b;
    }

    public void registerUndo(XJUndoDelegate delegate, JTextPane textPane) {
        undoEngine.registerUndo(new XJUndo(undoEngine, delegate), textPane);
    }

    public void performUndo() {
        XJUndo undo = getCurrentUndo();
        if(undo != null) {
            undo.performUndo();
        }
    }

    public void performRedo() {
        XJUndo undo = getCurrentUndo();
        if(undo != null) {
            undo.performRedo();
        }
    }

    public XJUndo getUndo(JTextPane textPane) {
        return undoEngine.getUndo(textPane);
    }

    public XJUndo getCurrentUndo() {
        return undoEngine.getCurrentUndo();
    }

    public boolean shouldDisplayMainMenuBar() {
        return true;
    }

    public boolean shouldAppearsInWindowMenu() {
        return false;
    }

    public void windowActivated() {

    }

    public void windowDeactivated() {

    }

    public void windowClosing() {

    }

    public void customizeFileMenu(XJMenu menu) {

    }

    public void customizeEditMenu(XJMenu menu) {

    }

    public void customizeWindowMenu(XJMenu menu) {

    }

    public void customizeHelpMenu(XJMenu menu) {

    }

    public void customizeMenuBar(XJMainMenuBar menubar) {

    }

    public void menuItemState(XJMenuItem item) {
        switch(item.getTag()) {
            case XJMainMenuBar.MI_NEW:
                item.setTitle(XJLocalizable.getXJString("New")+((XJApplication.shared().getDocumentExtensions().size()>1)?"...":""));
                break;
            case XJMainMenuBar.MI_UNDO:
            case XJMainMenuBar.MI_REDO:
                getMainMenuBar().menuUndoRedoItemState(undoEngine.getCurrentUndo());
                break;
        }
    }

    public void handleMenuEvent(XJMenu menu, XJMenuItem item) {
        switch(item.getTag()) {
            case XJMainMenuBar.MI_UNDO:
                performUndo();
                break;
            case XJMainMenuBar.MI_REDO:
                performRedo();
                break;
            case XJMainMenuBar.MI_CUT:
                performActionOnFocusedJComponent(DefaultEditorKit.cutAction);
                break;
            case XJMainMenuBar.MI_COPY:
                performActionOnFocusedJComponent(DefaultEditorKit.copyAction);
                break;
            case XJMainMenuBar.MI_PASTE:
                performActionOnFocusedJComponent(DefaultEditorKit.pasteAction);
                break;
            case XJMainMenuBar.MI_SELECT_ALL:
                performActionOnFocusedJComponent(DefaultEditorKit.selectAllAction);
                break;
        }
    }

    public void handleMenuSelected(XJMenu menu) {
    }

    public static void performActionOnFocusedJComponent(String action) {
        JComponent c = getFocusedJComponent();
        if(c != null)
            c.getActionMap().get(action).actionPerformed(null);
    }

    public static JComponent getFocusedJComponent() {
        Component c = KeyboardFocusManager.getCurrentKeyboardFocusManager().getPermanentFocusOwner();
        if(c instanceof JComponent)
            return (JComponent)c;
        else
            return null;
    }

    public Container getJavaContainer() {
        if(useDesktop) {
            return jInternalFrame;
        } else {
            return jFrame;
        }
    }

    private void restoreWindowBounds() {
        String name = autosaveName();
        if(name == null) return;

        Rectangle r = restoreWindowBoundsNewWay(name);
        if(r == null) {
            r = (Rectangle) XJApplication.shared().getPreferences().getObject(name, null);
        }

        if(r != null) {
            setPosition(r.x, r.y);
            setSize(r.width, r.height);
        }
    }

    public static final String PREF_WINDOWS_BOUNDS = "xjframe_bounds";

    private Rectangle restoreWindowBoundsNewWay(String name) {
        Map m = (Map) XJApplication.shared().getPreferences().getObject(PREF_WINDOWS_BOUNDS, null);
        if(m == null) return null;

        return (Rectangle) m.get(name);
    }

    protected void saveWindowBounds() {
        String name = autosaveName();
        if(name == null)
            return;

        Point pos = getLocation();
        Dimension s = getSize();
        Rectangle r = new Rectangle(pos.x, pos.y, s.width, s.height);
        XJPreferences prefs = XJApplication.shared().getPreferences();
        Map<String,Rectangle> docs = (Map<String,Rectangle>) prefs.getObject(PREF_WINDOWS_BOUNDS, null);
        if(docs == null) {
            docs = new HashMap<String, Rectangle>();
        }
        docs.put(name, r);
        prefs.setObject(PREF_WINDOWS_BOUNDS, docs);
    }

}
